#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>
#include "global.h"
#include "dynamic.h"
#include "heuristic.h"

char *readInt(char *str, int *res);
int loadData(FILE * fin);

State annealing();
State initState();
double initTemperature(State state);
State getNewState(State oldState, double temp);
double coolDown(double temp);
bool doInnerLoop(int iters);
bool doOuterLoop(double temp);
double getNormalRandom();

State getSomeRandomState();
State getNextRandomState(State state);
void log(State state);
void print(State state);

int instanceSize;
int instanceID;
int maxAllowedWeight;
int *itemCosts;
int *itemWeights;
FILE *flog = NULL;

/******************************************************************************
 * entry point
 */

int main(int argc, char* argv[]) {
		if(argc < 3) {
				printf("Jako parametr uvedte jmeno souboru\n");
				exit(1);
		}

		FILE *fin;
		if(fopen_s(&fin, argv[1], "r") != 0) {
				printf("Nelze otevrit %s\n", argv[3]);
				exit(1);
		}

		srand( (unsigned)time( NULL ) );

		int count = 1;
		readInt(argv[2], &count);

		time_t begin;
		time(&begin);
//		for(int i = 0; i < count; i++) {
				while(!feof(fin)) {
						if(loadData(fin) == 0) continue;
						char logName[30];
						sprintf_s(logName, 30, "output\\log%07d.log", instanceID);
						fopen_s(&flog, logName, "w");
						State d = dynamic();
						log(d);
						State best = annealing();
						fclose(flog);
				}
				rewind(fin);
//		}
		time_t ctime = time(NULL) - begin;
		printf("%f", (float) ctime / count);

		fclose(fin);
}

/******************************************************************************
 * loading data
 */

char *readInt(char *str, int *res) {
		*res = 0;

		while((str[0] == 0) || (str[0] == 10) || (str[0] == 13) || (str[0] == 32)) str++;

		while((str[0] != 0) && (str[0] != 10) && (str[0] != 13) && (str[0] != 32)) {
				if((str[0] >= '0') && (str[0] <= '9')) {
						*res = (*res * 10) + (str[0] - '0');
						str++;
				} else {
						printf("Neplatny znak ve vstupnim souboru: |%c|\n", str[0]);
						exit(1);
				}
		}
		return str;
}

int loadData(FILE * fin) {
		char line[1024] = "";
		char *act = line;
		fgets(line, 1024, fin);

		if((strlen(line) > 1024) || (strlen(line) < 5)) {
				return 0;
		}

		act = readInt(act, &instanceID);
		act = readInt(act, &instanceSize);
		if(instanceSize > (sizeof(unsigned long long) * 8)) {
				printf("Prilis mnoho polozek\n");
				exit(1);
		}
		act = readInt(act, &maxAllowedWeight);

		itemWeights = new int[instanceSize];
		itemCosts = new int[instanceSize];

		for(int i = 0; i < instanceSize; i++) {
				act = readInt(act, &itemWeights[i]);
				act = readInt(act, &itemCosts[i]);
		}

		return 1;
}

/******************************************************************************
 * simulated annealing
 */

State annealing() {
		State state = initState();
		double t = initTemperature(state);
		State maxState = state;
		log(state);

		bool terminate;
		do {
				terminate = true;
				int i = 0;
				do {
						State newState = getNewState(state, t);
						if(newState != state) {
								terminate = false;
								state = newState;
						}
						if(state.cost > maxState.cost) maxState = state;
						log(state);
						i++;
				} while(doInnerLoop(i));
				t = coolDown(t);
		} while(!terminate && doOuterLoop(t));
/*
		do {
				int i = 0;
				do {
						state = getNewState(state, t);
						if(state.cost > maxState.cost) maxState = state;
						log(state);
						i++;
				} while(doInnerLoop(i));
				t = coolDown(t);
		} while(doOuterLoop(t));
*/
		return maxState;
}

State initState() {
//		State s = heuristicTest();
//		State s = getSomeRandomState();
		State s = {0, 0, 0};
		return s;
}

double initTemperature(State state) {
		double sumCost = 0;
		double sumWeight = 0;
		for(int i = 0; i < instanceSize; i++) {
				sumCost += itemCosts[i];
				sumWeight += itemWeights[i];
		}
		return (sumCost / instanceSize) / (sumWeight / maxAllowedWeight) * 2;
}

State getNewState(State oldState, double temp) {
		State newState = getNextRandomState(oldState);
		int delta = newState.cost - oldState.cost;
		if(delta > 0) return newState;
		else {
				double r = getNormalRandom();
				if(r < exp(delta/temp)) return newState;
				else return oldState;
		} 
}

double coolDown(double temp) {
		return temp * 0.9;
}

bool doInnerLoop(int iters) {
		return iters < (instanceSize * 5);
}

bool doOuterLoop(double temp) {
		return temp > 0.1;
}

double getNormalRandom() {
		return rand() / (double) RAND_MAX;
}

/******************************************************************************
 * handling data
 */

State getSomeRandomState() {
		State newState = {0, 0, 0};
		do {
				for(int i = 0; i < instanceSize; i++) {
						if(getNormalRandom() < .1) {
								newState.value |= getLongLong(i);
								newState.cost += itemCosts[i];
								newState.weight += itemWeights[i];
						}
				}
		} while(newState.weight > maxAllowedWeight);
		return newState;
}

void exchange(State & state, int x) {
		unsigned long long item = getLongLong(x);
		if(state.value & item) {
				state.value &= ~item;
				state.cost -= itemCosts[x];
				state.weight -= itemWeights[x];
		} else {
				state.value |= item;
				state.cost += itemCosts[x];
				state.weight += itemWeights[x];
		}
}

State getNextRandomState(State state) {
		State newState;
		do {
				newState = state;
				int x = rand() % instanceSize;
				exchange(newState, x);
		} while(newState.weight > maxAllowedWeight);
		return newState;
}

void log(State state) {
		if(flog) {
				fprintf(flog, "%d\n", state.cost);
		}
}

void print(State state) {
		printf("%d %5d %4d/%4d ", instanceID, state.cost, state.weight, maxAllowedWeight);
		for(int i = 0; i < instanceSize; i++) {
				printf(" %d", ((state.value & getLongLong(i)) > 0));
		}
		printf("\n");
}
