#include <iostream>
#include <limits>
#include "dynamic.h"

using namespace std;

static State ***map;

State KNAP(unsigned long long V, int M, int last);
State trivialKNAP(unsigned long long V, int M);
unsigned long long reset(unsigned long long V, int i);
unsigned long long set(unsigned long long V, int i);
bool isTrivial (unsigned long long V, int M);

State dynamic() {
		map = new State **[instanceSize];
		for(int i = 0; i < instanceSize; i++) {
				map[i] = new State *[maxAllowedWeight];
				for(int j = maxAllowedWeight-1; j >= 0; j--) {
						map[i][j] = NULL;
				}
		}

		State s = KNAP(getLongLong(instanceSize) - 1, maxAllowedWeight, instanceSize-1);

		for(int i = 0; i < instanceSize; i++) {
				for(int j = maxAllowedWeight-1; j >= 0; j--) {
						delete map[i][j];
				}
				delete [] map[i];
		}
		delete map;

		return s;
}

State KNAP(unsigned long long V, int M, int last) {
    if(isTrivial(V, M)) return trivialKNAP(V, M);
		if(map[last][M-1]) {
				State res = {map[last][M-1]->value, map[last][M-1]->cost, map[last][M-1]->weight};
				return res;
		}

		State s1 = KNAP(reset(V, last), M, last-1);
		State s2 = KNAP(reset(V, last), M-itemWeights[last], last-1);

		s2.value = set(s2.value, last); s2.cost += itemCosts[last]; s2.weight += itemWeights[last];
		s1.value = reset(s1.value, last);

		if((s2.cost > s1.cost) && (s2.weight <= maxAllowedWeight)) {
				map[last][M-1] = new State(s2);
				return s2;
		} else {
				map[last][M-1] = new State(s1);
				return s1;
		}
}

unsigned long long reset(unsigned long long V, int i) {
		return V & ~getLongLong(i);
}

unsigned long long set(unsigned long long V, int i) {
		return V | getLongLong(i);
}

bool isTrivial (unsigned long long V, int M) {
    return (V == 0) || (M <= 0);
}

State trivialKNAP(unsigned long long V, int M) {
		if(M < 0) {
				State res = {0, INT_MIN, 0};
				return res;
		} else {
				State res = {0, 0, 0};
				return res;
		}
}
